// Copyright 2022 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package configuration

import (
	"github.com/kinbiko/jsonassert"
	"github.com/stretchr/testify/assert"
	"net/http"
	"net/http/httptest"
	"strings"
	"testing"
	"time"
)

type ConfigHttpStruct1 struct {
	A string `yaml:"A" json:"A" toml:"A" properties:"A"`
	B bool   `yaml:"B" json:"B" toml:"B" properties:"B"`
	C int    `yaml:"C" json:"C" toml:"C" properties:"C"`
}

func TestReadMeHttp(t *testing.T) {

	var config ConfigHttpStruct1
	s := New(config)

	req, err := http.NewRequest("GET", "/config", nil)
	if err != nil {
		t.Fatal(err)
	}

	req.Header.Set("Accept", "application/json")

	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(s.ServeHTTP)
	handler.ServeHTTP(rr, req)

	jsonassert.New(t).Assertf(rr.Body.String(),
		`{
  "A": "",
  "B": false,
  "C": 0
}
`)

}

func TestConfigurationMiddleware(t *testing.T) {

	req, err := http.NewRequest("GET", "/config", nil)
	if err != nil {
		t.Fatal(err)
	}

	nextHandler := http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		val := r.Context().Value(ContextConfigKey)
		if val == nil {
			t.Error("config not present")
			return
		}

		config, ok := val.(ConfigHttpStruct1)
		if !ok {
			t.Error("config not of correct type")
		}

		assert.Equal(t, config.A, "a")

	})

	var config ConfigHttpStruct1
	config.A = "a"
	s := New(config)

	s.Middleware(nextHandler).ServeHTTP(httptest.NewRecorder(), req)

}

func TestSettingServeHTTPJson(t *testing.T) {
	// Create a request to pass to our handler. We don't have any query parameters for now, so we'll
	// pass 'nil' as the third parameter.
	req, err := http.NewRequest("GET", "/config", nil)
	if err != nil {
		t.Fatal(err)
	}

	var config ConfigHttpStruct1
	s := New(config)

	req.Header.Set("Accept", "application/json")

	// We create a ResponseRecorder (which satisfies http.ResponseWriter) to record the response.
	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(s.ServeHTTP)

	// Our handlers satisfy http.Handler, so we can call their ServeHTTP method
	// directly and pass in our Request and ResponseRecorder.
	handler.ServeHTTP(rr, req)

	// Check the status code is what we expect.
	if status := rr.Code; status != http.StatusOK {
		t.Errorf("handler returned wrong status code: got %v want %v",
			status, http.StatusOK)
	}

	// Check the response body is what we expect.
	jsonassert.New(t).Assertf(rr.Body.String(),
		`{
  "A": "",
  "B": false,
  "C": 0
}
`)

}
func TestSettingServeHTTPYaml(t *testing.T) {
	// Create a request to pass to our handler. We don't have any query parameters for now, so we'll
	// pass 'nil' as the third parameter.
	req, err := http.NewRequest("GET", "/config", nil)
	if err != nil {
		t.Fatal(err)
	}

	req.Header.Set("Accept", "application/yaml")

	var config ConfigHttpStruct1
	s := New(config)

	// We create a ResponseRecorder (which satisfies http.ResponseWriter) to record the response.
	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(s.ServeHTTP)

	// Our handlers satisfy http.Handler, so we can call their ServeHTTP method
	// directly and pass in our Request and ResponseRecorder.
	handler.ServeHTTP(rr, req)

	// Check the status code is what we expect.
	if status := rr.Code; status != http.StatusOK {
		t.Errorf("handler returned wrong status code: got %v want %v",
			status, http.StatusOK)
	}

	// Check the response body is what we expect.
	yaml := `A: ""
B: false
C: 0
`
	if rr.Body.String() != yaml {
		t.Errorf("handler returned unexpected body: got %v want %v",
			rr.Body.String(), yaml)
	}

}

func TestSettingServeHTTPToml(t *testing.T) {

	// Create a request to pass to our handler. We don't have any query parameters for now, so we'll
	// pass 'nil' as the third parameter.
	req, err := http.NewRequest("GET", "/config", nil)
	if err != nil {
		t.Fatal(err)
	}

	req.Header.Set("Accept", "application/toml")

	var config ConfigHttpStruct1
	s := New(config)

	// We create a ResponseRecorder (which satisfies http.ResponseWriter) to record the response.
	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(s.ServeHTTP)

	// Our handlers satisfy http.Handler, so we can call their ServeHTTP method
	// directly and pass in our Request and ResponseRecorder.
	handler.ServeHTTP(rr, req)

	// Check the status code is what we expect.
	if status := rr.Code; status != http.StatusOK {
		t.Errorf("handler returned wrong status code: got %v want %v",
			status, http.StatusOK)
	}

	// Check the response body is what we expect.
	toml := `A = ''
B = false
C = 0
`
	if rr.Body.String() != toml {
		t.Errorf("handler returned unexpected body: got %v want %v",
			rr.Body.String(), toml)
	}

}

func TestSettingsHttpStatusNotAcceptable(t *testing.T) {
	req, err := http.NewRequest("GET", "/config", nil)
	if err != nil {
		t.Fatal(err)
	}

	req.Header.Set("Accept", "text/html")

	var config ConfigHttpStruct1

	s := New(config)

	// We create a ResponseRecorder (which satisfies http.ResponseWriter) to record the response.
	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(s.ServeHTTP)

	// Our handlers satisfy http.Handler, so we can call their ServeHTTP method
	// directly and pass in our Request and ResponseRecorder.
	handler.ServeHTTP(rr, req)

	// Check the status code is what we expect.
	if status := rr.Code; status != http.StatusNotAcceptable {
		t.Errorf("handler returned wrong status code: got %v want %v",
			status, http.StatusNotAcceptable)
	}
}

func TestSettingServeHTTPProperties(t *testing.T) {

	// Create a request to pass to our handler. We don't have any query parameters for now, so we'll
	// pass 'nil' as the third parameter.
	req, err := http.NewRequest("GET", "/config", nil)
	if err != nil {
		t.Fatal(err)
	}

	req.Header.Set("Accept", "application/properties")

	var config ConfigHttpStruct1

	s := New(config)

	// We create a ResponseRecorder (which satisfies http.ResponseWriter) to record the response.
	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(s.ServeHTTP)

	// Our handlers satisfy http.Handler, so we can call their ServeHTTP method
	// directly and pass in our Request and ResponseRecorder.
	handler.ServeHTTP(rr, req)

	// Check the status code is what we expect.
	if status := rr.Code; status != http.StatusOK {
		t.Errorf("handler returned wrong status code: got %v want %v",
			status, http.StatusOK)
	}

	// Check the response body is what we expect.
	// https://en.wikipedia.org/wiki/.properties
	// Each line in a .properties file normally stores a single property.
	// Several formats are possible for each line, including key=value,
	// key = value, key:value, and key value. Single-quotes or double-quotes
	// are considered part of the string. Trailing space is significant and
	// presumed to be trimmed as required by the consumer.

	r := rr.Body.String()
	// the order of the properties is not guaranteed
	assert.Contains(t, r, "A = ")
	assert.Contains(t, r, "B = false")
	assert.Contains(t, r, "C = 0")

}

func TestConfigurationServePostJson(t *testing.T) {

	var config ConfigHttpStruct1
	s := New(config)

	closeChan := make(chan bool)
	counter := 0

	var h ChangeHook
	h = &ChangeEventHandler{
		Callback: func(event ChangeEvent) {
			counter++
			closeChan <- true
		},
	}

	s.OnChange(h)

	// We create a ResponseRecorder (which satisfies http.ResponseWriter) to record the response.
	rr := httptest.NewRecorder()
	handler := http.HandlerFunc(s.ServeHTTP)

	// Our handlers satisfy http.Handler, so we can call their ServeHTTP method
	// directly and pass in our Request and ResponseRecorder.
	req, err := http.NewRequest("POST", "/config", strings.NewReader(`{
  "A": "hello",
  "B": true,
  "C": 1
}`))

	if err != nil {
		t.Fatal(err)
	}

	req.Header.Set("Content-Type", "application/json")
	handler.ServeHTTP(rr, req)

	// Check the status code is what we expect.
	if status := rr.Code; status != http.StatusOK {
		t.Errorf("handler returned wrong status code: got %v want %v",
			status, http.StatusOK)
	}

	config = s.Config()
	assert.Equal(t, "hello", config.A)
	assert.Equal(t, true, config.B)
	assert.Equal(t, 1, config.C)

	select {
	case <-closeChan:
	case <-time.After(time.Second * 5):
		t.Fatalf("Timeout")
	}

	if counter != 1 {
		t.Error("Expected 1 got ", counter)
	}

}
